@@ -1,7 +1,7 @@
 /*
  * utils.h - Misc utilities
  *
- * Copyright (C) 2013 - 2017, Max Lv <max.c.lv@gmail.com>
+ * Copyright (C) 2013 - 2018, Max Lv <max.c.lv@gmail.com>
  *
  * This file is part of the shadowsocks-libev.
  *
@@ -87,6 +87,40 @@ extern FILE *logfile;
 
 #else // not LIB_ONLY
 
+#ifdef __MINGW32__
+
+#define USE_TTY()
+#define USE_SYSLOG(ident, _cond)
+#define USE_LOGFILE(ident)
+#define TIME_FORMAT "%Y-%m-%d %H:%M:%S"
+#define LOGI(format, ...)                                    \
+    do {                                                     \
+        time_t now = time(NULL);                             \
+        char timestr[20];                                    \
+        strftime(timestr, 20, TIME_FORMAT, localtime(&now)); \
+        ss_color_info();                                     \
+        fprintf(stdout, " %s INFO: ", timestr);              \
+        ss_color_reset();                                    \
+        fprintf(stdout, format "\n", ## __VA_ARGS__);        \
+        fflush(stdout);                                      \
+    }                                                        \
+    while (0)
+
+#define LOGE(format, ...)                                     \
+    do {                                                      \
+        time_t now = time(NULL);                              \
+        char timestr[20];                                     \
+        strftime(timestr, 20, TIME_FORMAT, localtime(&now));  \
+        ss_color_error();                                     \
+        fprintf(stderr, " %s ERROR: ", timestr);              \
+        ss_color_reset();                                     \
+        fprintf(stderr, format "\n", ## __VA_ARGS__);         \
+        fflush(stderr);                                       \
+    }                                                         \
+    while (0)
+
+#else // not __MINGW32__
+
 #include <syslog.h>
 extern int use_tty;
 extern int use_syslog;
@@ -120,9 +154,11 @@ extern int use_syslog;
             if (use_tty) {                                                       \
                 fprintf(stdout, "\e[01;32m %s INFO: \e[0m" format "\n", timestr, \
                         ## __VA_ARGS__);                                         \
+                fflush(stdout);                                                  \
             } else {                                                             \
                 fprintf(stdout, " %s INFO: " format "\n", timestr,               \
                         ## __VA_ARGS__);                                         \
+                fflush(stdout);                                                  \
             }                                                                    \
         }                                                                        \
     }                                                                            \
@@ -139,18 +175,45 @@ extern int use_syslog;
             if (use_tty) {                                                        \
                 fprintf(stderr, "\e[01;35m %s ERROR: \e[0m" format "\n", timestr, \
                         ## __VA_ARGS__);                                          \
+                fflush(stderr);                                                   \
             } else {                                                              \
                 fprintf(stderr, " %s ERROR: " format "\n", timestr,               \
                         ## __VA_ARGS__);                                          \
+                fflush(stderr);                                                   \
             }                                                                     \
         } }                                                                       \
     while (0)
 
+#endif // if __MINGW32__
+
 #endif // if LIB_ONLY
 
 #endif // if __ANDROID__
 
+// Workaround for "%z" in Windows printf
+#ifdef __MINGW32__
+#define SSIZE_FMT "%Id"
+#define SIZE_FMT "%Iu"
+#else
+#define SSIZE_FMT "%zd"
+#define SIZE_FMT "%zu"
+#endif
+
+#ifdef __MINGW32__
+// Override Windows built-in functions
+#ifdef ERROR
+#undef ERROR
+#endif
+#define ERROR(s) ss_error(s)
+
+// Implemented in winsock.c
+void ss_error(const char *s);
+void ss_color_info(void);
+void ss_color_error(void);
+void ss_color_reset(void);
+#else
 void ERROR(const char *s);
+#endif
 
 char *ss_itoa(int i);
 int ss_isnumeric(const char *s);
@@ -167,10 +230,14 @@ void *ss_malloc(size_t size);
 void *ss_align(size_t size);
 void *ss_realloc(void *ptr, size_t new_size);
 
+int ss_is_ipv6addr(const char *addr);
+
 #define ss_free(ptr)     \
     do {                 \
         free(ptr);       \
         ptr = NULL;      \
     } while (0)
 
+char *get_default_conf(void);
+
 #endif // _UTILS_H
